@@ -1,13 +1,14 @@
-import asyncio
 import logging
+import sys
+import threading
 import types
 import time
 from datetime import timedelta, datetime
 from functools import partial
 from jsonmerge import merge
 from python_flux.flux_utilis import FluxUtils as fu
-from python_flux.subscribers import SSubscribe, SForeach
-
+from python_flux.subscribers import SSubscribe
+from concurrent.futures.thread import ThreadPoolExecutor
 
 class Flux(object):
 
@@ -33,6 +34,15 @@ class Flux(object):
         """
         return FMap(map_function, self)
 
+    def map_if(self, predicate, map_function):
+        """
+        Evalúa el predicado, si este es verdadero mapea el valor actual utilizando la función de mapeo.
+
+        :param predicate: función(valor, contexto) predicado que se evalúa.
+        :param map_function: función(valor, contexto) desde donde se obtendrá el valor a subsituir.
+        """
+        return FMapIf(predicate, map_function, self)
+
     def map_context(self, map_context_function):
         """
         Permite modificar el contexto de un flujo.
@@ -45,22 +55,24 @@ class Flux(object):
 
     def flat_map(self, flatmap_function):
         """
-        Permite modificar el valor de un flujo por lo elementos de otro flujo.
-        Recibe una función(valor, contexto) que se evaluará para cada elemento del flujo y retornará un flujo
-        cuyos elementos se usarán como nuevo flujo de datos.
+        Permite sustituir el valor de un flujo por elementos de otro flujo.
+        Recibe una función(valor, contexto) que se evaluará para cada elemento del flujo original y retornará un
+        flujo cuyos elementos se usarán como nuevo origen del flujo de datos.
 
         flatmap_function: función(valor, contexto) desde donde se obtendrá el valor a subsituir.
         """
         return FFlatMap(flatmap_function, self)
 
-    def do_on_next(self, on_next):
+    def do_on_next(self, on_next, on_error=fu.default_action, exec_async=True):
         """
         Ejecuta par cada elemento del flujo una acción de forma asincrónica. No afecta a los valores
         del flujo.
 
-        :param on_next: función(valor, contexto) que se ejecutará para cada valor del flujo
+        :param on_next: función(valor, contexto) se ejecuta para cada valor del flujo
+        :param on_error: función(ex, contexto) se ejecuta en caso de error
+        :param exec_async: Booleano que indica si las funciones se ejecutarán de forma asíncrona
         """
-        return FDoOnNext(on_next, self)
+        return FDoOnNext(on_next, on_error, exec_async, self)
 
     def delay_ms(self, delay_ms, predicate=fu.default_predicate):
         """
@@ -72,6 +84,16 @@ class Flux(object):
         """
         return FDelayMs(delay_ms, predicate, self)
 
+    def chunks(self, n):
+        """
+        Emite un valor construido como una lista de como máximo n valores obtenidos del flujo padre.
+        Si algún valor mientras se construye el chunk da error se emiten los elementos recolectados
+        y luego se propaga el error.
+
+        :param n: Cantidad de elementos recolectados antes de emitir el evento de lista.
+        """
+        return FChunks(n, self)
+
     def take(self, n):
         """
         Corta la ejecución del flujo luego de n elementos procesados.
@@ -90,26 +112,38 @@ class Flux(object):
         """
         return FTakeDuringTimeDelta(n, self)
 
-    def log(self, message=lambda v, c: str(v), level=logging.INFO):
+    def log(self, build_log_message=lambda v, c: str(v), build_error_message=lambda e, c: str(e), level=logging.INFO):
         """
         Recibe una función(valor, contexto) que retorna un mensaje de texto que será logueado con el
         nivel de log indicado en level.
         El logueo se hace de forma asincrónica y no afecta al flujo.
 
-        :param message: Función que dado un valor y contexto retorna un mensaje a loguear.
+        :param build_log_message: Función que dado un valor y contexto retorna un mensaje a loguear
+        :param build_error_message: Función que dada una excepción y el contexto retorna el mensaje a loguear como ERROR
         :param level: Nivel de logueo que se usará
         """
         def log_function(l, msg, v, c):
-            if l is logging.ERROR:
-                logging.error(msg(v, c))
-            elif l is logging.WARNING or l is logging.WARN:
-                logging.warning(msg(v, c))
+            message = msg(v, c)
+            if l is logging.WARNING or l is logging.WARN:
+                logging.warning(message)
             elif l is logging.INFO:
-                logging.info(msg(v, c))
+                logging.info(message)
             elif l is logging.DEBUG:
-                logging.debug(msg(v, c))
+                logging.debug(message)
 
-        return FDoOnNext(partial(log_function, level, message), self)
+        def log_error(msg, e, c):
+            message = msg(e, c)
+            logging.error(message)
+
+        return FDoOnNext(partial(log_function, level, build_log_message), partial(log_error, build_error_message), True, self)
+
+    def parallel(self, pool_size=5):
+        """
+        Ejecuta la obtención de elementos del stream en paralelo
+
+        :param pool_size: Pool de hilos a utilizar
+        """
+        return FParallel(pool_size, self)
 
     def on_error_resume(self, resume=fu.default_error_resume, exceptions=[Exception]):
         """
@@ -135,47 +169,49 @@ class Flux(object):
         """
         return FOnErrorRetry(retries, exceptions, delay_ms, self)
 
-    def subscribe(self, context={}):
+    def subscribe(self, context={}, skip_error=False):
         """
         Crea un objeto iterable a partir del flujo. Si se itera sobre este objeto se obtendrán
         los valores del flujo.
+        :param skip_error: Ignora errores al obtener los valores desde el flujo
         :param context: Contexto inicial para el flujo
         :return: Objeto iterable
         """
-        return SSubscribe(context, self)
+        return SSubscribe(skip_error, context, self)
 
-    def foreach(self, on_success=fu.default_success, on_error=fu.default_error, context={}):
+    def foreach(self, on_success=fu.default_success, on_error=fu.default_error, on_finish=fu.default_finish, context={}):
         """
         Itera sobre los elementos del flujo e invoca a funciones on_success y on_error dependiendo
         el estado del flujo.
         :param on_success: función(valor, contexto) se invoca si el flujo procesa correctamente un valor
-        :param on_error:  función(ex, contexto) se invoca si hay un error en el flujo.
+        :param on_error:  función(ex, valor, contexto) se invoca si hay un error en el flujo.
                           Esto no corta el procesamiento a menos que se lance una excepción en el método
+        :param on_finish: función(contexto) se invoca al finalizar el flujo
         :param context: Contexto inicial para el flujo
         """
-        flux = self.subscribe(context)
+        flux = self.subscribe(context, True)
         while True:
             try:
                 value = next(flux)
                 fu.try_or(partial(on_success), value, context)
             except StopIteration:
+                fu.try_or(partial(on_finish), context)
                 return
             except Exception as ex:
-                _, e = fu.try_or(partial(on_error), ex, context)
-                if e is not None:
-                    raise e
+                _, e = fu.try_or(partial(on_error), ex, value, context)
 
-    def to_list(self, context={}):
+    def to_list(self, context={}, skip_error=True):
         """
         Itera sobre los elementos del flujo y los retorna todos dentro de una lista.
         :param context: contexto inicial para el flujo
+        :param skip_error: Ignora errores al obtener los valores desde el flujo
         :return: Lista de elementos
         """
-        return list(iter(SSubscribe(context, self)))
+        return list(iter(self.subscribe(context, skip_error)))
 
     def collect(self, init=lambda c: {}, reduce=lambda v, a: a, context={}):
         """
-        Permite resumir en un sólo objeto el procesamiento de todos los elementos del flujo.
+        Permite agrupar en un sólo objeto el resultado del procesamiento de todos los elementos del flujo.
         Se inicializa un acumulador a través de la función init(contexto) y luego para cada elemento del flujo
         se invoca la función reduce(valor, acumulador) que procesa el valor y retorna un nuevo valor del acumulador.
         :param init: función(contexto) Retorna valor inicial del acumulador
@@ -184,7 +220,7 @@ class Flux(object):
         :return: Acumulador
         """
         acum = init(context)
-        flux = self.subscribe(context)
+        flux = self.subscribe(context, False)
         while True:
             try:
                 value = next(flux)
@@ -204,7 +240,9 @@ class Stream(Flux):
         self.upstream.prepare_next()
 
     def next(self, context):
+        t = time.process_time()
         value, e, ctx = self.upstream.next(context)
+        elapsed_time = time.process_time() - t
         return value, e, ctx
 
 
@@ -218,16 +256,42 @@ class FFilter(Stream):
         value, e, ctx = super(FFilter, self).next(context)
         if e is not None:
             return value, e, ctx
-        valid, e = fu.try_or(partial(self.predicate), value, context)
+
+        valid, e = fu.try_or(partial(self.predicate), value, ctx)
         while e is None and not valid:
-            _, e = fu.try_or(partial(self.on_mismatch), value, context)
+            _, e = fu.try_or(partial(self.on_mismatch), value, ctx)
             if e is not None:
                 return value, e, ctx
             super(FFilter, self).prepare_next()
-            value, e, ctx = super(FFilter, self).next(context)
+            value, e, new_ctx = super(FFilter, self).next(ctx)
             if e is not None:
                 return value, e, ctx
-            valid, e = fu.try_or(partial(self.predicate), value, context)
+            ctx = merge(ctx, new_ctx)
+            valid, e = fu.try_or(partial(self.predicate), value, ctx)
+        return value, e, ctx
+
+
+class FMapIf(Stream):
+    def __init__(self, p, f, flux):
+        super().__init__(flux)
+        self.predicate = p
+        self.function = f
+
+    def next(self, context):
+        value, e, ctx = super(FMapIf, self).next(context)
+        if e is not None:
+            return value, e, ctx
+
+        valid, e = fu.try_or(partial(self.predicate), value, ctx)
+        if e is not None:
+            return value, e, ctx
+
+        if valid:
+            v, e = fu.try_or(partial(self.function), value, ctx)
+            if e is not None:
+                return value, e, ctx
+            return v, e, ctx
+
         return value, e, ctx
 
 
@@ -266,6 +330,38 @@ class FTakeDuringTimeDelta(Stream):
             return value, StopIteration(), ctx
 
 
+class FChunks(Stream):
+    def __init__(self, count, flux):
+        super().__init__(flux)
+        self.count = count
+        self.buffer = []
+        self.stop_iteration = False
+
+    def prepare_next(self):
+        if not self.stop_iteration:
+            super(FChunks, self).prepare_next()
+
+    def next(self, context):
+        if self.stop_iteration:
+            return self.buffer, StopIteration(), context
+        ctx = context
+        while len(self.buffer) < self.count:
+            value, e, new_ctx = super(FChunks, self).next(ctx)
+            if e is not None:
+                self.stop_iteration = True
+                if len(self.buffer) == 0:
+                    return self.buffer, StopIteration(), ctx
+                buf = self.buffer.copy()
+                self.buffer = []
+                return buf, None, ctx
+            self.buffer.append(value)
+            super(FChunks, self).prepare_next()
+            ctx = merge(ctx, new_ctx)
+        buf = self.buffer.copy()
+        self.buffer = []
+        return buf, None, ctx
+
+
 class FDelayMs(Stream):
     def __init__(self, delay, predicate, flux):
         super().__init__(flux)
@@ -276,13 +372,77 @@ class FDelayMs(Stream):
         value, e, ctx = super(FDelayMs, self).next(context)
         if e is not None:
             return value, e, ctx
-        valid, e = fu.try_or(partial(self.predicate), value, context)
+        valid, e = fu.try_or(partial(self.predicate), value, ctx)
         if e is not None:
             return value, e, ctx
         if valid:
             time.sleep(self.delay / 1000)
         return value, e, ctx
 
+from random import randint
+class _Register:
+
+    def __init__(self, id):
+        self.id = id
+        self.value = None
+        self.e = None
+        self.ctx = None
+
+    def execute(self, prepare_next, next, context):
+        self.value, self.e, self.ctx = next(context)
+        prepare_next()
+        time.sleep(randint(1, 3))
+        sys.stdout.write(f'get: {self.id} {self.value}\n')
+        sys.stdout.flush()
+
+
+class FParallel(Stream):
+    def __init__(self, pool, flux):
+        super().__init__(flux)
+        self.pool_size = pool
+        self.jobs = {}
+        self.pool = {}
+        self.ids = 0
+        self.waiting_to_stop = False
+
+    def next(self, context):
+        ctx = context
+        while True:
+            if not self.waiting_to_stop:
+                if len(self.pool) < self.pool_size:
+                    reg = _Register(self.ids)
+                    t = threading.Thread(target=reg.execute, args=[super(FParallel, self).prepare_next, super(FParallel, self).next, context])
+                    self.jobs[self.ids] = reg
+                    self.pool[self.ids] = t
+                    self.ids = self.ids + 1
+                    t.start()
+            jobs = self.pool.copy()
+            for i, t in jobs.items():
+                try:
+                    t.join(0.001)
+                    if t.is_alive():
+                        continue
+                    else:
+                        print(f'pop {i}')
+                        self.pool.pop(i)
+                        r = self.jobs.pop(i)
+                        ctx = merge(ctx, r.ctx)
+                        if isinstance(r.e, StopIteration):
+                            print('StopIteration')
+                            self.waiting_to_stop = True
+                            if len(self.pool) == 0:
+                                return None, StopIteration(), ctx
+                            else:
+                                continue
+                        if self.waiting_to_stop:
+                            if len(self.pool) == 0:
+                                return None, StopIteration(), ctx
+                        print(f'Return {i} {r.value}')
+                        return r.value, r.e, ctx
+                except Exception as e:
+                    print(str(e))
+                    return None, e, ctx
+
 
 class FMap(Stream):
     def __init__(self, func, flux):
@@ -320,6 +480,12 @@ class FFlatMap(Stream):
         self.function = func
         self.current = None
 
+    def prepare_next(self):
+        if self.current is None:
+            super(FFlatMap, self).prepare_next()
+        else:
+            self.current.prepare_next()
+
     def next(self, context):
         ctx = context.copy()
         while True:
@@ -330,19 +496,20 @@ class FFlatMap(Stream):
                 func, e = fu.try_or(partial(self.function), value, ctx)
                 if e is not None:
                     return None, e, ctx
+                fgen = func
                 if isinstance(func, types.GeneratorType):
                     from python_flux.producers import PFromGenerator
                     fgen = PFromGenerator(func)
-                else:
-                    fgen = func
-                self.current = fgen.subscribe(ctx)
-            try:
-                v = next(self.current)
-                return v, None, ctx
-            except StopIteration:
+                self.current = fgen
+                self.current.prepare_next()
+
+            cur_value, cur_e, cur_ctx = self.current.next(ctx)
+            if cur_e is None:
+                return cur_value, cur_e, cur_ctx
+            if isinstance(cur_e, StopIteration):
                 self.current = None
-            except Exception as ex:
-                return None, ex, ctx
+                super(FFlatMap, self).prepare_next()
+            ctx = merge(ctx, cur_ctx)
 
 
 class FOnErrorResume(Stream):
@@ -358,7 +525,7 @@ class FOnErrorResume(Stream):
         if isinstance(e, StopIteration):
             return None, e, ctx
         if any([isinstance(e, ex) for ex in self.exceptions]):
-            v, e = fu.try_or(partial(self.on_error_resume), e, ctx)
+            v, e = fu.try_or(partial(self.on_error_resume), e, value, ctx)
             if e is not None:
                 return value, e, ctx
             value = v
@@ -389,16 +556,25 @@ class FOnErrorRetry(Stream):
 
 
 class FDoOnNext(Stream):
-    async def __async_coroutine(self, f, value, context):
-        f(value, context)
 
-    def __init__(self, func, flux):
+    def __init__(self, on_next, on_error, exec_async, flux):
         super().__init__(flux)
-        self.function = func
+        self.on_next = on_next
+        self.on_error = on_error
+        self.execute_async = exec_async
 
     def next(self, context):
         value, e, ctx = super(FDoOnNext, self).next(context)
         if e is not None:
+            if self.execute_async:
+                t = threading.Thread(target=self.on_error, args=(e, ctx,))
+                t.start()
+            else:
+                self.on_error(e, ctx)
             return value, e, ctx
-        asyncio.run(self.__async_coroutine(self.function, value, ctx))
+        if self.execute_async:
+            t = threading.Thread(target=self.on_next, args=(value, ctx,))
+            t.start()
+        else:
+            self.on_next(value, ctx)
         return value, e, ctx
